#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include "pri.h"

#define GETBYTE(input)        fgetc((FILE*)(input))
#include "unrle.c"


static FILE *g_input;   /* input file */


/* pri_readheader()
 * Helper function that reads a bitmap header and checks whether it is valid
 */
int pri_readheader(FILE *file, PRHDR *pri)
{
    unsigned char header[PRHEADERSIZE];

    if (fread(header, 1, PRHEADERSIZE, file) < PRHEADERSIZE)
        return 0;

    pri->size   = PRHDR_SIZE(header);
    pri->id     = PRHDR_ID(header);
    pri->layout = PRHDR_LAYOUT(header);
    pri->width  = PRHDR_WIDTH(header);
    pri->height = PRHDR_HEIGHT(header);
    pri->depth  = PRHDR_DEPTH(header);

    if (pri->size == 0 || pri->id != PRSIGNATURE)
        return 0;

    return 1;
}

int pri_readextheader(FILE *file, const PRHDR *pri, PREXTHDR *exthdr)
{
    unsigned char header[PREXTHEADERSIZE];

    /* read the optional extended header */
    if ((pri->layout & PRLAYOUT_EXTHDR) == 0)
        return 0;

    fread(header, 1, PREXTHEADERSIZE, file);
    exthdr->delay = PREXTHDR_DELAY(header);
    exthdr->dx    = PREXTHDR_DX(header);
    exthdr->dy    = PREXTHDR_DY(header);

    return 1;
}

void pri_readdata(FILE *file, const PRHDR *pri, unsigned char *data)
{
    struct rlestat stat = { 0 };
    long size = pri->height * ((pri->width + 7) / 8);

    while (size-- > 0)
        *data++ = unrle(file, &stat);
}

/* This helper function copies pixel data into a Device Dependent Bitmap (DDB).
 * It is needed because the DDB format requires that scan lines are aligned
 * on 16-bit boundaries, meaning that the size of a scan line must be padded
 * to a multiple of two bytes. The poly-raster format does not use scan-line
 * alignement or padding.
 */
void CopyDDBitmap(HBITMAP hbmp, unsigned char *pixels, int x, int y, int width, int height)
{
    BITMAP bm;
    unsigned char *newpix;

    GetObject(hbmp, sizeof bm, &bm);
    newpix = malloc(bm.bmHeight * bm.bmWidthBytes);
    if (newpix) {
        int srcscan = (width + 7) / 8;  /* convert to # bytes per scan line */
        int line;
        x /= 8;
        GetBitmapBits(hbmp, height * bm.bmWidthBytes, newpix);
        for (line = 0; line < height; line++)
            memcpy(newpix + (line + y) * bm.bmWidthBytes + x, pixels + line * srcscan, srcscan);
        SetBitmapBits(hbmp, bm.bmHeight * bm.bmWidthBytes, newpix);
        free(newpix);
    } /* if */
}

LRESULT CALLBACK ViewProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    static int g_width, g_height;
    static PRHDR g_pri;
    static PREXTHDR g_exthdr;
    static unsigned char *g_image = NULL;
    static HBITMAP g_hbmp = NULL;

    switch (msg) {
    case WM_CREATE: {
        RECT rect;
        int width,height;

        /* read the image or the first frame of an animation */
        for ( ;; ) {
            if (!pri_readheader(g_input, &g_pri)) {
                MessageBox(hwnd, "No supported bitmap type found", NULL, MB_OK);
                DestroyWindow(hwnd);
                return 1;
            }
            if (g_pri.layout == 0 && g_pri.depth == 1)
                break;
            fseek(g_input, g_pri.size - sizeof g_pri, SEEK_CUR);
        }
        g_image = malloc(PRBITMAPSIZE(g_pri));
        if (g_image == NULL) {
            MessageBox(hwnd, "Insufficient memory", NULL, MB_OK);
            DestroyWindow(hwnd);
            return 1;
        }
        pri_readdata(g_input, &g_pri, g_image);
        /* save the bitmap size, as new frames overwrite the header */
        g_width = g_pri.width;
        g_height = g_pri.height;

        /* create a DDB from the pixel data, with the appropriate size */
        g_hbmp = CreateBitmap(g_width, g_height, 1, g_pri.depth, NULL);
        CopyDDBitmap(g_hbmp, g_image, 0, 0, g_width, g_height);

        /* resize and centre the window */
        SetRect(&rect, 0, 0, g_width, g_height);
        AdjustWindowRectEx(&rect, WS_POPUPWINDOW | WS_CAPTION, FALSE, WS_EX_CLIENTEDGE | WS_EX_TOOLWINDOW);
        width=GetSystemMetrics(SM_CXSCREEN);
        height=GetSystemMetrics(SM_CYSCREEN);
        SetWindowPos(hwnd, 0,
                     (width - (rect.right-rect.left))/2,
                     (height - (rect.bottom-rect.top))/2,
                     rect.right - rect.left, rect.bottom - rect.top,
                     SWP_NOACTIVATE | SWP_NOZORDER);

        /* see if there are more frames in the file, if so set a timer for
         * animation
         */
        while (pri_readheader(g_input, &g_pri)) {
            if (g_pri.layout == PRLAYOUT_EXTHDR && g_pri.depth == 1) {
                pri_readextheader(g_input, &g_pri, &g_exthdr);
                SetTimer(hwnd, 1, g_exthdr.delay, NULL);
                break;
            }
            fseek(g_input, g_pri.size - sizeof g_pri, SEEK_CUR);
        }

        break;
    } /* case */

    case WM_DESTROY:
        if (g_input != NULL)
            fclose(g_input);
        if (g_hbmp != NULL)
            DeleteObject(g_hbmp);
        if (g_image != NULL)
            free(g_image);
        KillTimer(hwnd, 1);
        PostQuitMessage(0);
        break;

    case WM_PAINT: {
        PAINTSTRUCT ps;
        HDC hdcMem;
        HBITMAP hbmp;
        BeginPaint(hwnd, &ps);
        hdcMem = CreateCompatibleDC(ps.hdc);
        hbmp = SelectObject(hdcMem, g_hbmp);
        BitBlt(ps.hdc, 0, 0, g_width, g_height, hdcMem, 0, 0, SRCCOPY);
        SelectObject(hdcMem, hbmp);
        DeleteDC(hdcMem);
        EndPaint(hwnd, &ps);
        break;
    } /* case */

    case WM_KEYDOWN:
        if (wParam == VK_ESCAPE)
            DestroyWindow(hwnd);
        break;

    case WM_TIMER:
        /* the header and extended header for this frame have already been read;
         * now read the remaining part of the image and copy it into the DDB
         * that already exists
         */
        pri_readdata(g_input, &g_pri, g_image);
        CopyDDBitmap(g_hbmp, g_image, g_exthdr.dx, g_exthdr.dy,
                     g_pri.width, g_pri.height);
        /* redraw the window */
        InvalidateRect(hwnd, NULL, FALSE);
        UpdateWindow(hwnd);
        /* find the next frame, optionally loop back to the beginning */
        if ((g_pri.layout & PRLAYOUT_LOOPBACK) != 0)
            rewind(g_input);
        for ( ;; ) {
            if (!pri_readheader(g_input, &g_pri)) {
                KillTimer(hwnd, 1);
                break;
            } /* if */
            if ((g_pri.layout & ~PRLAYOUT_LOOPBACK) == PRLAYOUT_EXTHDR && g_pri.depth == 1) {
                pri_readextheader(g_input, &g_pri, &g_exthdr);
                SetTimer(hwnd, 1, g_exthdr.delay, NULL);
                break;
            }
            fseek(g_input, g_pri.size - sizeof g_pri, SEEK_CUR);
        }
        break;

    default:
        return DefWindowProc(hwnd, msg, wParam, lParam);
    } /* switch */

    return 0L;
}

int main(int argc, char *argv[])
{
    WNDCLASS wc;
    MSG msg;
    const char *caption;

    if (argc != 2)
        fprintf(stderr, "usage: %s <input>\n", argv[0]);

    /* open the file */
    g_input = fopen(argv[1], "rb");
    if (g_input == NULL) {
        MessageBox(NULL, "Input file cannot be opened, or it is not found", NULL, MB_OK);
        return 1;
    }

    caption = strrchr(argv[1], '\\');
    if (caption != NULL)
        caption++;
    else
        caption = argv[1];

    /* create a window for the image (with a dummy size) */
    memset(&wc, 0, sizeof wc);
    wc.lpfnWndProc = ViewProc;
    wc.hInstance = GetModuleHandle(NULL);
    wc.hbrBackground = (HBRUSH)(COLOR_WINDOW + 1);
    wc.lpszClassName = "pri:view";
    RegisterClass(&wc);
    CreateWindowEx(WS_EX_CLIENTEDGE | WS_EX_TOOLWINDOW | WS_EX_APPWINDOW,
                   "pri:view", caption,
                   WS_VISIBLE | WS_POPUPWINDOW | WS_CAPTION,
                   CW_USEDEFAULT, CW_USEDEFAULT, 0, 0,
                   NULL, NULL, GetModuleHandle(NULL), NULL);

    /* the message loop */
    while (GetMessage(&msg, NULL, 0, 0)) {
        TranslateMessage(&msg);
        DispatchMessage(&msg);
    } /* while */

    return 0;
}
